/***************************************************************************
                          mididriver.cpp  -  description
                             -------------------
    begin                : Sun Jun 9 2002
    copyright            : (C) 2002 by Morten Hustveit
    email                : mortehu@ifi.uio.no
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <unistd.h>
#include <dlfcn.h>
#include <dirent.h>
#include <errno.h>
#ifdef HAVE_ALSA
#  include <alsa/asoundlib.h>
#endif

#include <qstring.h>

#include <kdebug.h>

#include "mididriver.h"
#include "configuration.h"

class RawPort : public MIDIDriver::Port
{
public:

  RawPort(const char* name)
  {
    _fd = open(name, O_RDWR);

    if(_fd == -1)
      kdWarning() << "Failed to open MIDI port read/write: " << strerror(errno) << endl;
  }

  ~RawPort()
  {
    ::close(_fd);
  }

  void put(int data)
  {
    _buffer.push_back(data);
  }

  void putRT(int data)
  {
    if(status && data == status)
      return;

    if(data >= 0x80 && data <= 0xEF)
      status = data;
    else if(data >= 0xF0 && data <= 0xF7)
      status = 0;

    unsigned char cdata = (unsigned char) data;

    write(_fd, &cdata, 1);
  }

  void clockStep()
  {
    for(uint j = 0; j < _buffer.size(); ++j)
      putRT(_buffer[j]);

    _buffer.clear();
  }

  bool empty()
  {
    struct pollfd pfd;

    pfd.fd = _fd;
    pfd.events = POLLIN;
    pfd.revents = 0;

    poll(&pfd, 1, 0);

    return !(pfd.revents & POLLIN);
  }

  void drain()
  {
    _buffer.clear();
  }

  QValueList<struct pollfd> pollfds()
  {
    QValueList<struct pollfd> ret;

    struct pollfd pfd;

    pfd.fd = _fd;
    pfd.events = POLLIN;
    pfd.revents = 0;

    ret.push_back(pfd);

    return ret;
  }

  int get()
  {
    unsigned char data;

    read(_fd, &data, 1);

    return (int) data;
  }

  bool isOpen() const
  {
    return _fd != -1;
  }

protected:

  int                       _fd;
  QValueList<unsigned char> _buffer;
};


#ifdef HAVE_ALSA
class ALSAPort : public MIDIDriver::Port
{
public:

  ALSAPort(Configuration::PortInfo info)
    : localPort(-1),
      _info(info)
  {
    if(!seq)
      return;

    unsigned int caps = 0;

    if(_info.input)
      caps |= SND_SEQ_PORT_CAP_READ | SND_SEQ_PORT_CAP_SUBS_READ;

    if(_info.output)
      caps |= SND_SEQ_PORT_CAP_WRITE | SND_SEQ_PORT_CAP_SUBS_WRITE;

    localPort = snd_seq_create_simple_port(seq, 0, caps, SND_SEQ_PORT_TYPE_MIDI_GENERIC);

    if(_info.input)
      snd_seq_connect_from(seq, localPort, _info.client, _info.port);

    if(_info.output)
      snd_seq_connect_to(seq, localPort, _info.client, _info.port);

    snd_seq_ev_clear(&eventRT);
    snd_seq_ev_set_direct(&eventRT);
    snd_seq_ev_set_source(&eventRT, localPort);
    snd_seq_ev_set_dest(&eventRT, _info.client, _info.port);
  }

  ~ALSAPort()
  {
    if(localPort == -1)
      return;

    if(_info.input)
      snd_seq_disconnect_from(seq, localPort, _info.client, _info.port);

    if(_info.output)
      snd_seq_disconnect_to(seq, localPort, _info.client, _info.port);
  }

  void put(int data)
  {
    _outBuffer.push_back(data);
  }

  void clockStep()
  {
    for(uint j = 0; j < _outBuffer.size(); ++j)
      putRT(_outBuffer[j]);

    _outBuffer.clear();
  }

  bool empty()
  {
    get_events();

    return _inBuffer.empty();
  }

  void drain()
  {
    // Note: this drains all open ALSA ports, not just this one
    snd_seq_drain_output(seq);
  }

  void putRT(int data)
  {
    if(data >= 0x80)
    {
      _status = data;

      _datapos = 0;

      if(_status <= 0xF0)
        return;
    }

    snd_seq_event_t event = eventRT;

    if((_status & 0xF0) == MIDI_NOTE_ON(0))
    {
      if(_datapos == 1)
      {
        snd_seq_ev_set_noteon(&event, _status & 0x0F, _data[0] /* note */, data /* velocity */);

        snd_seq_event_output_direct(seq, &event);

        _datapos = 0;
      }
      else
        _data[_datapos++] = data;
    }
    else if((_status & 0xF0) == MIDI_NOTE_OFF(0))
    {
      if(_datapos == 1)
      {
        snd_seq_ev_set_noteoff(&event, _status & 0x0F, _data[0] /* note */, data /* velocity */);

        snd_seq_event_output_direct(seq, &event);

        _datapos = 0;
      }
      else
        _data[_datapos++] = data;
    }
    else if((_status & 0xF0) == MIDI_CONTROL(0))
    {
      if(_datapos == 1)
      {
        snd_seq_ev_set_controller(&event, _status & 0x0F, _data[0], data);

        snd_seq_event_output_direct(seq, &event);
      }
      else
        _data[_datapos++] = data;
    }
    else if((_status & 0xF0) == MIDI_PATCH_CHANGE(0))
    {
      snd_seq_ev_set_pgmchange(&event, _status & 0x0F, data);

      snd_seq_event_output_direct(seq, &event);
    }
    else if(_status == MIDI_PLAY)
    {
      event.type = SND_SEQ_EVENT_START;

      snd_seq_event_output_direct(seq, &event);
    }
    else if(_status == MIDI_STOP)
    {
      event.type = SND_SEQ_EVENT_STOP;

      snd_seq_event_output_direct(seq, &event);
    }
    else if(_status == MIDI_CLOCK)
    {
      event.type = SND_SEQ_EVENT_CLOCK;

      snd_seq_event_output_direct(seq, &event);
    }
    else
    {
      kdDebug() << "Unknown event type " << _status << endl;
    }
  }

  QValueList<struct pollfd> pollfds()
  {
    QValueList<struct pollfd> ret;

    int count = snd_seq_poll_descriptors_count(seq, POLLIN);

    if(!count)
      return ret;

    struct pollfd* pollfds = new struct pollfd[count];

    snd_seq_poll_descriptors(seq, pollfds, count, POLLIN);

    for(int i = 0; i < count; ++i)
      ret.push_back(pollfds[i]);

    delete [] pollfds;

    return ret;
  }

  int get()
  {
    if(!_inBuffer.empty())
    {
      int data = _inBuffer.front();

      _inBuffer.pop_front();

      return data;
    }

    int data = _inBuffer.front();
    _inBuffer.pop_front();
    
    return data;
  }

  bool isOpen() const
  {
    return localPort != -1;
  }

  static void init();
  static void close();

  static Configuration::PortInfos portInfos();

private:

  void get_events()
  {
    int count = snd_seq_event_input_pending(seq, 1);

    if(count < 0)
      return;

    while(count--)
    {
      snd_seq_event_t* event;

      snd_seq_event_input(seq, &event);

      switch(event->type)
      {
      case SND_SEQ_EVENT_CLOCK:

        _inBuffer.push_back(MIDI_CLOCK);

        break;

      case SND_SEQ_EVENT_NOTEON:

        _inBuffer.push_back(MIDI_NOTE_ON(event->data.note.channel));
        _inBuffer.push_back(event->data.note.note);
        _inBuffer.push_back(event->data.note.velocity);

        break;

      case SND_SEQ_EVENT_NOTEOFF:

        _inBuffer.push_back(MIDI_NOTE_OFF(event->data.note.channel));
        _inBuffer.push_back(event->data.note.note);
        _inBuffer.push_back(event->data.note.velocity);

        break;

      case SND_SEQ_EVENT_START:

        _inBuffer.push_back(MIDI_PLAY);

        break;

      case SND_SEQ_EVENT_STOP:

        _inBuffer.push_back(MIDI_STOP);

        break;

      case SND_SEQ_EVENT_SENSING: // "I'm here" event sent mostly by Roland gear
      case SND_SEQ_EVENT_PORT_SUBSCRIBED: // Sent when port is first connected
      case SND_SEQ_EVENT_PORT_START:      // Sent when port is first connected

        break;

      default:

        kdDebug() << "Received event of unknown type " << (uint) event->type << endl;
      }

      snd_seq_free_event(event);
    }
  }

  int _status;
  int _data[2];
  int _datapos;
  int localPort;
  Configuration::PortInfo _info;
  QValueList<unsigned char> _inBuffer;
  QValueList<unsigned char> _outBuffer;

  snd_seq_event_t eventRT;

  static void* sohandle;
  static int   queue;

  typedef int          (*t_snd_seq_open)(snd_seq_t** handle, const char* name, int streams, int mode);
  typedef void         (*t_snd_seq_client_info_set_client)(snd_seq_client_info_t *info, int client);
  typedef int          (*t_snd_seq_query_next_client)(snd_seq_t *handle, snd_seq_client_info_t *info);
  typedef void         (*t_snd_seq_port_info_set_client)(snd_seq_port_info_t *info, int client);
  typedef void         (*t_snd_seq_port_info_set_port)(snd_seq_port_info_t *info, int port);
  typedef int          (*t_snd_seq_query_next_port)(snd_seq_t *handle, snd_seq_port_info_t *info);
  typedef unsigned int (*t_snd_seq_port_info_get_capability)(const snd_seq_port_info_t *info);
  typedef size_t       (*t_snd_seq_client_info_sizeof)(void);
  typedef size_t       (*t_snd_seq_port_info_sizeof)(void);
  typedef int          (*t_snd_seq_client_info_get_client)(const snd_seq_client_info_t *info);
  typedef const char*  (*t_snd_seq_port_info_get_name)(const snd_seq_port_info_t *info);
  typedef int          (*t_snd_seq_port_info_get_port)(const snd_seq_port_info_t *info);
  typedef int          (*t_snd_seq_drain_output)(snd_seq_t *handle);
  typedef int          (*t_snd_seq_poll_descriptors_count)(snd_seq_t *seq, short events);
  typedef int          (*t_snd_seq_poll_descriptors)(snd_seq_t *seq, struct pollfd *pfds, unsigned int space, short events);
  typedef int          (*t_snd_seq_create_simple_port)(snd_seq_t *handle, const char *name, unsigned int caps, unsigned int type);
  typedef int          (*t_snd_seq_alloc_queue)(snd_seq_t *handle);
  typedef int          (*t_snd_seq_free_queue)(snd_seq_t *handle, int q);
  typedef int          (*t_snd_seq_close)(snd_seq_t *handle);
  typedef int          (*t_snd_seq_control_queue)(snd_seq_t *handle, int q, int type, int value, snd_seq_event_t *ev);
  typedef int          (*t_snd_seq_connect_from)(snd_seq_t *handle, int my_port, int src_client, int src_port);
  typedef int          (*t_snd_seq_connect_to)(snd_seq_t *handle, int my_port, int dest_client, int dest_port);
  typedef int          (*t_snd_seq_disconnect_from)(snd_seq_t *handle, int my_port, int src_client, int src_port);
  typedef int          (*t_snd_seq_disconnect_to)(snd_seq_t *handle, int my_port, int dest_client, int dest_port);
  typedef int          (*t_snd_seq_event_output)(snd_seq_t *handle, snd_seq_event_t *ev);  
  typedef int          (*t_snd_seq_event_input_pending)(snd_seq_t *seq, int fetch_sequencer);
  typedef int          (*t_snd_seq_event_input)(snd_seq_t *handle, snd_seq_event_t **ev);
  typedef int          (*t_snd_seq_free_event)(snd_seq_event *ev);
  typedef int          (*t_snd_seq_event_output_direct)(snd_seq_t *handle, snd_seq_event_t *ev);
  typedef int          (*t_snd_seq_queue_tempo_malloc)(snd_seq_queue_tempo_t **ptr);
  typedef void         (*t_snd_seq_queue_tempo_free)(snd_seq_queue_tempo_t *ptr);
  typedef void         (*t_snd_seq_queue_tempo_set_tempo)(snd_seq_queue_tempo_t *info, unsigned int tempo);
  typedef void         (*t_snd_seq_queue_tempo_set_ppq)(snd_seq_queue_tempo_t *info, int ppq);
  typedef int          (*t_snd_seq_set_queue_tempo)(snd_seq_t *handle, int q, snd_seq_queue_tempo_t *tempo);
  
  static t_snd_seq_open                     snd_seq_open;
  static t_snd_seq_client_info_set_client   snd_seq_client_info_set_client;
  static t_snd_seq_query_next_client        snd_seq_query_next_client;
  static t_snd_seq_port_info_set_client     snd_seq_port_info_set_client;
  static t_snd_seq_port_info_set_port       snd_seq_port_info_set_port;
  static t_snd_seq_query_next_port          snd_seq_query_next_port;
  static t_snd_seq_port_info_get_capability snd_seq_port_info_get_capability;
  static t_snd_seq_client_info_sizeof       snd_seq_client_info_sizeof;
  static t_snd_seq_port_info_sizeof         snd_seq_port_info_sizeof;
  static t_snd_seq_client_info_get_client   snd_seq_client_info_get_client;
  static t_snd_seq_port_info_get_name       snd_seq_port_info_get_name;
  static t_snd_seq_port_info_get_port       snd_seq_port_info_get_port;
  static t_snd_seq_drain_output             snd_seq_drain_output;
  static t_snd_seq_poll_descriptors_count   snd_seq_poll_descriptors_count;
  static t_snd_seq_poll_descriptors         snd_seq_poll_descriptors;
  static t_snd_seq_create_simple_port       snd_seq_create_simple_port;
  static t_snd_seq_alloc_queue              snd_seq_alloc_queue;
  static t_snd_seq_control_queue            snd_seq_control_queue;
  static t_snd_seq_free_queue               snd_seq_free_queue;
  static t_snd_seq_close                    snd_seq_close;
  static t_snd_seq_connect_from             snd_seq_connect_from;
  static t_snd_seq_connect_to               snd_seq_connect_to;
  static t_snd_seq_disconnect_from          snd_seq_disconnect_from;
  static t_snd_seq_disconnect_to            snd_seq_disconnect_to;
  static t_snd_seq_event_output             snd_seq_event_output;
  static t_snd_seq_event_input_pending      snd_seq_event_input_pending;
  static t_snd_seq_event_input              snd_seq_event_input;
  static t_snd_seq_free_event               snd_seq_free_event;
  static t_snd_seq_event_output_direct      snd_seq_event_output_direct;
  static t_snd_seq_queue_tempo_malloc       snd_seq_queue_tempo_malloc;
  static t_snd_seq_queue_tempo_free         snd_seq_queue_tempo_free;
  static t_snd_seq_queue_tempo_set_tempo    snd_seq_queue_tempo_set_tempo;
  static t_snd_seq_queue_tempo_set_ppq      snd_seq_queue_tempo_set_ppq;
  static t_snd_seq_set_queue_tempo          snd_seq_set_queue_tempo;
    
  static snd_seq_t* seq;
};
#endif

class aRtsPort : public MIDIDriver::Port
{
public:

  ~aRtsPort()
  {
  }

  void put(int /*data*/)
  {
  }

  void clockStep()
  {
  }

  bool empty()
  {
    return true;
  }

  void drain()
  {
  }

  void putRT(int /*data*/)
  {
  }

  QValueList<struct pollfd> pollfds()
  {
    return QValueList<struct pollfd>();
  }

  int get()
  {
    return 0;
  }

  bool isOpen() const
  {
    return false;
  }
};

int                          MIDIDriver::_reconfPipe[2] = { -1 };
QValueList<unsigned char>    MIDIDriver::_buffer;
QPtrVector<MIDIDriver::Port> MIDIDriver::_ports;

void MIDIDriver::init()
{
  pipe(_reconfPipe);

#ifdef HAVE_ALSA
  ALSAPort::init();
#endif
}

void MIDIDriver::close()
{
  unsigned char msg = 1;
  ::write(_reconfPipe[1], &msg, 1);

  _ports.clear();

#ifdef HAVE_ALSA
  ALSAPort::close();
#endif
}

void MIDIDriver::reconfigure(const Configuration* config)
{
  unsigned char msg = 0;
  write(_reconfPipe[1], &msg, 1);

  _ports.setAutoDelete(true);
  _ports.clear();

  uint n = 0;

  for(uint i = 0; i < config->portCount(); ++i)
  {
    Configuration::PortInfo::Type type = config->port(i).type();

    Port* newPort;

    if(type == Configuration::PortInfo::Raw)
    {
      newPort = new RawPort(config->port(i).deviceFile());
    }
#ifdef HAVE_ALSA
    else if(type == Configuration::PortInfo::ALSA)
    {
      newPort = new ALSAPort(config->port(i).info());
    }
#endif
    else if(type == Configuration::PortInfo::aRts)
    {
      newPort = new aRtsPort();
    }
    else
    {
      kdError() << "Unsupported port type "
                << static_cast<int>(config->port(i).type())
                << endl;

      continue;
    }

    if(newPort->isOpen())
    {
      _ports.resize(n + 1);
      _ports.insert(n++, newPort);
    }
    else
    {
      delete newPort;
    }
  }

  write(_reconfPipe[1], &msg, 1);
}

void MIDIDriver::put(int port, int data)
{
  _ports[port]->put(data);
}

void MIDIDriver::broadcast(int data)
{
  for(uint i = 0; i < _ports.size(); ++i)
    _ports[i]->put(data);
}

void MIDIDriver::broadcastRT(int data)
{
  for(uint i = 0; i < _ports.size(); ++i)
    _ports[i]->putRT(data);
}

void MIDIDriver::clockStep()
{
  for(uint i = 0; i < _ports.size(); ++i)
    _ports[i]->clockStep();
}

void MIDIDriver::drain()
{
  for(uint i = 0; i < _ports.size(); ++i)
    _ports[i]->drain();
}

void MIDIDriver::putRT(int port, int data)
{
  if(port < 0 || static_cast<uint>(port) >= _ports.size())
  {
    kdWarning() << "MIDIDriver::putRT: Port " << port << " out of range."
                << endl;

    return;
  }

  _ports[port]->putRT(data);
}

void MIDIDriver::setBPM(double /*bpm*/)
{
  // XXX: implementation needed
}

double MIDIDriver::bpm()
{
  // XXX: implementation needed

  return 0;
}

int MIDIDriver::get(int& port)
{
  if(_reconfPipe[0] == -1) // MIDI driver closed
    return MIDI_STOP;
 
  port = 0; // XXX
  if(!_buffer.empty())
  {
    int data = _buffer.front();

    _buffer.pop_front();

    return data;
  }

  for(;;)
  {
    QValueList<struct pollfd> pollfds;
    QValueList<int> pollfdSep; // Position of last pollfd for port
   
    struct pollfd pfd;
    pfd.fd = _reconfPipe[0];
    pfd.events = POLLIN;
    pfd.revents = 0;

    pollfds.push_back(pfd);

    for(uint i = 0; i < _ports.size(); ++i)
    {
      pollfds += _ports[i]->pollfds();
      pollfdSep.push_back(pollfds.size());
    }

    // Create a C array suitable for passing to poll()
    struct pollfd* pfds = new struct pollfd[pollfds.size()];
    for(uint i = 0; i < pollfds.size(); ++i)
      pfds[i] = pollfds[i];
    
    poll(pfds, pollfds.size(), -1);

    // Check whether a reconfiguration has happened
    if(pfds[0].revents & POLLIN)
    {
      delete [] pfds;

      // reconfigure() writes a '0' before reconfiguration a '0' after
      // close() writes a '1'
      unsigned char msg;
      ::read(_reconfPipe[0], &msg, 1);

      if(msg == 1)
      {
        ::close(_reconfPipe[0]);
        ::close(_reconfPipe[1]);

        _reconfPipe[0] = _reconfPipe[1] = -1;

        return MIDI_STOP;
      }

      ::read(_reconfPipe[0], &msg, 1);

      continue;
    }

    int pfdi = 1;

    for(uint i = 0; i < _ports.size(); ++i)
    {
      bool data = false;

      while(pfdi < pollfdSep[i])
      {
        if(pfds[pfdi++].revents & POLLIN)
          data = true;
      }

      if(data)
      {
        while(!_ports[i]->empty())
          _buffer.push_back(_ports[i]->get());
      }
    }

    delete [] pfds;

    if(!_buffer.empty())
    {
      int data = _buffer.front();

      _buffer.pop_front();

      return data;
    }
  }
}

int MIDIDriver::portCount()
{
  return _ports.size();
}

#ifdef HAVE_ALSA
ALSAPort::t_snd_seq_open                     ALSAPort::snd_seq_open;
ALSAPort::t_snd_seq_client_info_set_client   ALSAPort::snd_seq_client_info_set_client;
ALSAPort::t_snd_seq_query_next_client        ALSAPort::snd_seq_query_next_client;
ALSAPort::t_snd_seq_port_info_set_client     ALSAPort::snd_seq_port_info_set_client;
ALSAPort::t_snd_seq_port_info_set_port       ALSAPort::snd_seq_port_info_set_port;
ALSAPort::t_snd_seq_query_next_port          ALSAPort::snd_seq_query_next_port;
ALSAPort::t_snd_seq_port_info_get_capability ALSAPort::snd_seq_port_info_get_capability;
ALSAPort::t_snd_seq_client_info_sizeof       ALSAPort::snd_seq_client_info_sizeof;
ALSAPort::t_snd_seq_port_info_sizeof         ALSAPort::snd_seq_port_info_sizeof;
ALSAPort::t_snd_seq_client_info_get_client   ALSAPort::snd_seq_client_info_get_client;
ALSAPort::t_snd_seq_port_info_get_name       ALSAPort::snd_seq_port_info_get_name;
ALSAPort::t_snd_seq_port_info_get_port       ALSAPort::snd_seq_port_info_get_port;
ALSAPort::t_snd_seq_drain_output             ALSAPort::snd_seq_drain_output;
ALSAPort::t_snd_seq_poll_descriptors_count   ALSAPort::snd_seq_poll_descriptors_count;
ALSAPort::t_snd_seq_poll_descriptors         ALSAPort::snd_seq_poll_descriptors;
ALSAPort::t_snd_seq_create_simple_port       ALSAPort::snd_seq_create_simple_port;
ALSAPort::t_snd_seq_alloc_queue              ALSAPort::snd_seq_alloc_queue;
ALSAPort::t_snd_seq_control_queue            ALSAPort::snd_seq_control_queue;
ALSAPort::t_snd_seq_free_queue               ALSAPort::snd_seq_free_queue;
ALSAPort::t_snd_seq_close                    ALSAPort::snd_seq_close;
ALSAPort::t_snd_seq_connect_from             ALSAPort::snd_seq_connect_from;
ALSAPort::t_snd_seq_connect_to               ALSAPort::snd_seq_connect_to;
ALSAPort::t_snd_seq_disconnect_from          ALSAPort::snd_seq_disconnect_from;
ALSAPort::t_snd_seq_disconnect_to            ALSAPort::snd_seq_disconnect_to;
ALSAPort::t_snd_seq_event_output             ALSAPort::snd_seq_event_output;
ALSAPort::t_snd_seq_event_input_pending      ALSAPort::snd_seq_event_input_pending;
ALSAPort::t_snd_seq_event_input              ALSAPort::snd_seq_event_input;
ALSAPort::t_snd_seq_free_event               ALSAPort::snd_seq_free_event;
ALSAPort::t_snd_seq_event_output_direct      ALSAPort::snd_seq_event_output_direct;
ALSAPort::t_snd_seq_queue_tempo_malloc       ALSAPort::snd_seq_queue_tempo_malloc;
ALSAPort::t_snd_seq_queue_tempo_free         ALSAPort::snd_seq_queue_tempo_free;
ALSAPort::t_snd_seq_queue_tempo_set_tempo    ALSAPort::snd_seq_queue_tempo_set_tempo;
ALSAPort::t_snd_seq_queue_tempo_set_ppq      ALSAPort::snd_seq_queue_tempo_set_ppq;
ALSAPort::t_snd_seq_set_queue_tempo          ALSAPort::snd_seq_set_queue_tempo;


void*      ALSAPort::sohandle = 0;
int        ALSAPort::queue = 0;
snd_seq_t* ALSAPort::seq = 0;

void ALSAPort::init()
{
  if(seq || sohandle)
    return;

  sohandle = dlopen("libasound.so.2", RTLD_NOW);

  if(!sohandle)
  {
    kdWarning() << dlerror() << "'', ALSA disabled." << endl;
 
    return;
  }

  // We can test the symbols for NULL instead of testing for errors, since no 
  // usable functions can be NULL-pointers.
  if(!(snd_seq_open =                     (t_snd_seq_open)                     dlsym(sohandle, "snd_seq_open"))
  || !(snd_seq_client_info_set_client =   (t_snd_seq_client_info_set_client)   dlsym(sohandle, "snd_seq_client_info_set_client"))
  || !(snd_seq_query_next_client =        (t_snd_seq_query_next_client)        dlsym(sohandle, "snd_seq_query_next_client"))
  || !(snd_seq_port_info_set_client =     (t_snd_seq_port_info_set_client)     dlsym(sohandle, "snd_seq_port_info_set_client"))
  || !(snd_seq_port_info_set_port =       (t_snd_seq_port_info_set_port)       dlsym(sohandle, "snd_seq_port_info_set_port"))
  || !(snd_seq_query_next_port =          (t_snd_seq_query_next_port)          dlsym(sohandle, "snd_seq_query_next_port"))
  || !(snd_seq_port_info_get_capability = (t_snd_seq_port_info_get_capability) dlsym(sohandle, "snd_seq_port_info_get_capability"))
  || !(snd_seq_client_info_sizeof =       (t_snd_seq_client_info_sizeof)       dlsym(sohandle, "snd_seq_client_info_sizeof"))
  || !(snd_seq_port_info_sizeof =         (t_snd_seq_port_info_sizeof)         dlsym(sohandle, "snd_seq_port_info_sizeof"))
  || !(snd_seq_client_info_get_client =   (t_snd_seq_client_info_get_client)   dlsym(sohandle, "snd_seq_client_info_get_client"))
  || !(snd_seq_port_info_get_name =       (t_snd_seq_port_info_get_name)       dlsym(sohandle, "snd_seq_port_info_get_name"))
  || !(snd_seq_port_info_get_port =       (t_snd_seq_port_info_get_port)       dlsym(sohandle, "snd_seq_port_info_get_port"))
  || !(snd_seq_drain_output =             (t_snd_seq_drain_output)             dlsym(sohandle, "snd_seq_drain_output"))
  || !(snd_seq_poll_descriptors_count =   (t_snd_seq_poll_descriptors_count)   dlsym(sohandle, "snd_seq_poll_descriptors_count"))
  || !(snd_seq_poll_descriptors =         (t_snd_seq_poll_descriptors)         dlsym(sohandle, "snd_seq_poll_descriptors"))
  || !(snd_seq_create_simple_port =       (t_snd_seq_create_simple_port)       dlsym(sohandle, "snd_seq_create_simple_port"))
  || !(snd_seq_alloc_queue =              (t_snd_seq_alloc_queue)              dlsym(sohandle, "snd_seq_alloc_queue"))
  || !(snd_seq_control_queue =            (t_snd_seq_control_queue)            dlsym(sohandle, "snd_seq_control_queue"))
  || !(snd_seq_free_queue =               (t_snd_seq_free_queue)               dlsym(sohandle, "snd_seq_free_queue"))
  || !(snd_seq_close =                    (t_snd_seq_close)                    dlsym(sohandle, "snd_seq_close"))
  || !(snd_seq_connect_from =             (t_snd_seq_connect_from)             dlsym(sohandle, "snd_seq_connect_from"))
  || !(snd_seq_connect_to =               (t_snd_seq_connect_to)               dlsym(sohandle, "snd_seq_connect_to"))
  || !(snd_seq_disconnect_from =          (t_snd_seq_disconnect_from)          dlsym(sohandle, "snd_seq_disconnect_from"))
  || !(snd_seq_disconnect_to =            (t_snd_seq_disconnect_to)            dlsym(sohandle, "snd_seq_disconnect_to"))
  || !(snd_seq_event_output =             (t_snd_seq_event_output)             dlsym(sohandle, "snd_seq_event_output"))
  || !(snd_seq_event_input_pending =      (t_snd_seq_event_input_pending)      dlsym(sohandle, "snd_seq_event_input_pending"))
  || !(snd_seq_event_input =              (t_snd_seq_event_input)              dlsym(sohandle, "snd_seq_event_input"))
  || !(snd_seq_free_event =               (t_snd_seq_free_event)               dlsym(sohandle, "snd_seq_free_event"))
  || !(snd_seq_event_output_direct =      (t_snd_seq_event_output_direct)      dlsym(sohandle, "snd_seq_event_output_direct"))
  || !(snd_seq_queue_tempo_malloc =       (t_snd_seq_queue_tempo_malloc)       dlsym(sohandle, "snd_seq_queue_tempo_malloc"))
  || !(snd_seq_queue_tempo_free =         (t_snd_seq_queue_tempo_free)         dlsym(sohandle, "snd_seq_queue_tempo_free"))
  || !(snd_seq_queue_tempo_set_tempo =    (t_snd_seq_queue_tempo_set_tempo)    dlsym(sohandle, "snd_seq_queue_tempo_set_tempo"))
  || !(snd_seq_queue_tempo_set_ppq =      (t_snd_seq_queue_tempo_set_ppq)      dlsym(sohandle, "snd_seq_queue_tempo_set_ppq"))
  || !(snd_seq_set_queue_tempo =          (t_snd_seq_set_queue_tempo)          dlsym(sohandle, "snd_seq_set_queue_tempo")))
  {
    kdWarning() << dlerror() << "'', ALSA disabled." << endl;

    ::dlclose(sohandle);

    sohandle = 0;

    return;
  }

  if(snd_seq_open(&seq, "default", SND_SEQ_OPEN_DUPLEX, 0) < 0)
  {
    kdWarning() << "Failed to open ALSA sequencer, ALSA disabled." << endl;

    ::dlclose(sohandle);

    sohandle = 0;

    return;
  }

  queue = snd_seq_alloc_queue(seq);

#if 0
  snd_seq_queue_tempo_t* queue_tempo;

  snd_seq_queue_tempo_malloc(&queue_tempo);
  int tempo = (int)(6e7 / ((double)140 * (double)128) * (double)128);
  snd_seq_queue_tempo_set_tempo(queue_tempo, tempo);
  snd_seq_queue_tempo_set_ppq(queue_tempo, 128);
  snd_seq_set_queue_tempo(seq, queue, queue_tempo);
  snd_seq_queue_tempo_free(queue_tempo);

  snd_seq_start_queue(seq, queue, 0);
#endif
}

void ALSAPort::close()
{
  if(!seq)
    return;

  snd_seq_free_queue(seq, queue);

  snd_seq_close(seq);

  if(sohandle)
    dlclose(sohandle);
}
#endif // HAVE_ALSA

Configuration::PortInfos MIDIDriver::portInfos()
{
  Configuration::PortInfos portInfos;

  // Scan for Raw MIDI ports

  DIR* dir = opendir("/dev");

  struct dirent* ent;

  while(0 != (ent = readdir(dir)))
  {
    if(ent->d_name == strstr(ent->d_name, "midi"))
    {
      Configuration::PortInfo portInfo;

      portInfo.name = QString("/dev/") + ent->d_name;

      int fd;

      if(-1 != (fd = open(portInfo.name, O_RDWR)))
      {
        portInfo.type = Configuration::PortInfo::Raw;
        portInfo.input = true;
        portInfo.output = true;

        portInfos.push_back(portInfo);

        ::close(fd);
      }
    }
  }

  closedir(dir);

#ifdef HAVE_ALSA
  portInfos += ALSAPort::portInfos();
#endif

  return portInfos;
}

#ifdef HAVE_ALSA
Configuration::PortInfos ALSAPort::portInfos()
{
  Configuration::PortInfos portInfos;

  if(seq)
  { 
    snd_seq_client_info_t* cinfo;
    snd_seq_port_info_t* pinfo;

    snd_seq_client_info_alloca(&cinfo);
    snd_seq_port_info_alloca(&pinfo);
    snd_seq_client_info_set_client(cinfo, -1);

    while(snd_seq_query_next_client(seq, cinfo) >= 0)
    {
      snd_seq_port_info_set_client(pinfo, snd_seq_client_info_get_client(cinfo));
      snd_seq_port_info_set_port(pinfo, -1);

      while(snd_seq_query_next_port(seq, pinfo) >= 0)
      {
#define perm_ok(pinfo,bits) ((snd_seq_port_info_get_capability(pinfo) & (bits)) == (bits))

        if(snd_seq_port_info_get_capability(pinfo) & SND_SEQ_PORT_CAP_NO_EXPORT)
          continue;

        Configuration::PortInfo portInfo;

        portInfo.input = perm_ok(pinfo, SND_SEQ_PORT_CAP_READ | SND_SEQ_PORT_CAP_SUBS_READ);
        portInfo.output = perm_ok(pinfo, SND_SEQ_PORT_CAP_WRITE | SND_SEQ_PORT_CAP_SUBS_WRITE);

        if(!portInfo.input && !portInfo.output)
          continue;

        portInfo.type = Configuration::PortInfo::ALSA;
        portInfo.client = snd_seq_client_info_get_client(cinfo);
        portInfo.port = snd_seq_port_info_get_port(pinfo);
        portInfo.name = QString("%1 (%2) (%3:%4)")
          .arg(snd_seq_port_info_get_name(pinfo))
          .arg(portInfo.input ? (portInfo.output ? "Input/Output" : "Input") : "Output")
          .arg(portInfo.client)
          .arg(portInfo.port);
         
        portInfos.push_back(portInfo);
      }
    }
  }

  return portInfos;
}
#endif // HAVE_ALSA

