/***************************************************************************
                          sanetrackerdoc.h  -  description
                             -------------------
    begin                : Sat Jun  8 15:33:59 CEST 2002
    copyright            : (C) 2002 by Morten Hustveit
    email                : mortehu@ifi.uio.no
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef SANETRACKERDOC_H
#define SANETRACKERDOC_H

#include <qobject.h>
#include <qstring.h>
#include <qptrlist.h>
#include <qmap.h>

#include <kurl.h>

#include "midifile.h"

class Configuration;

/**
 * The SanetrackerDoc class provides methods for processing, reading and
 * writing Sanetracker song files, which is a format 100% compatible with the
 * Standard MIDI Format (SMF).
 */
class SanetrackerDoc : public QObject
{
Q_OBJECT

public:

  /**
   * Constructs an empty Sanetracker document, using @p config as
   * configuration.
   */
  SanetrackerDoc(QObject* parent, const Configuration& config);

  /**
   * Destroys a Sanetracker document.
   */
  ~SanetrackerDoc();

  /**
   * The Pattern class is used to represent one measure of a music score.
   *
   * Unlike traditional trackers, Sanetracker uses one set of patterns for each
   * device.  Each pattern may contain an unlimited number of tracks.
   */
  class Pattern
  {
  public:

    /**
     * Constructs an empty pattern.
     */
    Pattern();

    /**
     * Copies a pattern.
     */
    Pattern(const Pattern& pattern);

    /**
     * Destroys a pattern.
     */
    ~Pattern();

    /**
     * Copies a pattern.
     */
    Pattern& operator=(const Pattern& pattern);

    /**
     * The Track class represents one channel of a track.
     *
     * Notes occuring in a track are played one at a time; if you need several
     * simultanous notes, you need to use multiple tracks.
     */
    class Track
    {
    public:

      /**
       * Constructs an empty track.
       */
      Track();

      /**
       * Copies a track.
       */
      Track(const Track& track);

      /**
       * Destroys a track.
       */
      ~Track();

      /**
       * Copies a track.
       */
      Track& operator=(const Track& track);

      /**
       * The Cell class represents a note on or a note off event.
       */
      class Cell
      {
      public:

        /**
         * Constructs an empty cell.
         */
        Cell()
          : _data(-1)
        {
        }

        /**
         * Returns TRUE if this cell is a note, and FALSE otherwise. The
         * alternatives are stop cells and empty cells.
         *
         * @see setNote
         */
        bool isNote() const
        {
          return _data >= 0;
        }

        /**
         * Returns TRUE if a this cell is empty, and FALSE otherwise.
         *
         * @see clear
         */
        bool isEmpty() const
        {
          return _data == -1;
        }

        /**
         * Returns TRUE if this cell is a stop (note off) marker, and FALSE
         * otherwise. The alternatvies are note cells and empty cells.
         *
         * @see setStop
         */
        bool isStop() const
        {
          return _data == -2;
        }

        /**
         * Returns the MIDI note this cell contains, i.e. a value in the range
         * from 0 to 127, inclusive.
         *
         * @see setNote
         */
        unsigned char note() const
        {
          return _data & 0x7F;
        }

        /**
         * Returns the relativity of this cell's velocity value.
         *
         * -1 means this cell's velocity should be subtracted from the current
         * value, 1 means it should be added to the current value, and 0 means
         * it should replace the current value.
         *
         * @see setRelativity
         * @see velocity
         */
        int relativity() const
        {
          return ((_data >> 14) & 0x3) - 1;
        }

        /**
         * Returns the velocity this cell contains.  The value is in the range
         * from 0 to 127, inclusive.  You must take the relatvity into account
         * when using this value.
         *
         * @see relativity
         * @see setVelocity
         */
        uint velocity() const
        {
          return (_data >> 7) & 0x7F;
        }

        /**
         * Clears the contents of this cell.
         *
         * @see isEmpty
         */
        void clear()
        {
          _data = -1;
        }

        /**
         * Makes this cell a stop (note off) marker.
         *
         * @see isStop
         */
        void setStop()
        {
          _data = -2;
        }

        /**
         * Sets the MIDI note for this cell.
         *
         * @see note
         */
        void setNote(unsigned char note)
        {
          if(!isNote())
            _data = note;
          else
            _data = (_data & ~0x7F) | note;
        }

        /**
         * Sets the relativity of this cell's velocity value. See
         * @ref relativity for valid values of @p dir.
         *
         * @see relativity
         * @see velocity
         */
        void setRelativity(int dir)
        {
          if(isNote())
            _data = (_data & ~0xC000) | ((dir + 1) << 14);
        }

        /**
         * Sets the (absolute or relative) velocity for this cell.
         *
         * @see velocity
         * @see relativity
         */
        void setVelocity(uint value)
        {
          if(isNote())
            _data = (_data & ~0x3F80) | (value << 7);
        }

      protected:

        int _data;
      };

      /**
       * Returns the cell on row @p index in this track. If @p index is out of
       * range, new empty cells are automatically created, and all references
       * to cells in this track are invalidated.
       */
      Cell& operator[](uint index);

      /**
       * Returns the cell on row @p index in this track. Unlike the non-const
       * []-operator, no new cells will be added if @p index
       * is out of range.
       */
      const Cell& operator[](uint index) const;

    private:

      friend class Pattern;

      static const Cell _emptyCell;

      Cell* _cells;
      uint  _cellCount;
    };

    /**
     * Returns the patch used by this pattern. This is index is used in
     * conjunction with the @ref Configuration class, and does not
     * neccesarily correspond to a MIDI patch number.
     *
     * @see setPatch
     */
    uint patch() const
    {
      return _patch;
    }

    /**
     * Sets this pattern's patch to @p patch.
     *
     * @see patch
     */
    void setPatch(uint patch)
    {
      _patch = patch;
    }

    /**
     * Returns the track in column @p index of this pattern. If @p index is
     * out of range, new empty cells are automatically created, and all
     * references to columns in this pattern are invalidated.
     */
    Track& operator[](uint index);

    /**
     * Returns the track in column @p index of this pattern. Unlike the
     * non-const []-operator, no new columns will be added if @p index
     * is out of range.
     */
    const Track& operator[](uint index) const;

    /**
     * Removes the track with the index @p index. Invalidates all references
     * to tracks in this pattern.
     */
    void remove(uint index);

    /**
     * Returns the amount of tracks in this pattern.
     */
    uint size() const
    {
      return _trackCount;
    }

  protected:

    static const Track _emptyTrack;

    uint   _patch;
    Track* _tracks;
    uint   _trackCount;
  };

  typedef MIDIFile::MTrk::Event Event;

  /**
   * The EventList class provides a sorted list of events, and functions for
   * dealing with such lists efficiently.
   */
  class EventList
  {
  public:

    /**
     * Constructs an empty event list.
     */
    EventList();

    /**
     * Constructs a list of the events from @p events that has the timestamp
     * @p time.
     */

    EventList(const EventList& events, uint time);

    /**
     * Adds the event @p event to the list.
     */
    EventList& operator+=(const Event& event);

    /**
     * Returns the amount of events in this list.
     */
    uint size() const;

    /**
     * Returns the event with index @p index.
     */
    Event& operator[](uint index);

    /**
     * Returns the event with index @p index.
     */
    const Event& operator[](uint index) const;

    /**
     * Removes the event with index @p index.
     */
    void remove(uint index);

  protected:

    QValueList<Event> _events;
  };

  /**
   * The Device class represents a single MIDI device.
   *
   * It contains all the events and patterns associated with the device.
   */
  class Device
  {
  public:

    /**
     * Creates a device associated with the configuration @p config, port index
     * @p portIndex, device ID @p deviceID, configured name @p configName and
     * configured type @p configType.
     */
    Device(const Configuration& config, uint portIndex, uint deviceID,
      const QString& configName, const QString& configType);

    /**
     * Returns the index of the port this device is associated with.
     */
    uint portIndex() const
    {
      return _portIndex;
    }

    /**
     * Returns the MIDI device ID of this device.  The value is in the range
     * from 0 to 15, inclusive.
     */
    uint deviceID() const
    {
      return _deviceID;
    }

    /**
     * Returns the configured name of this device.
     */
    const QString& configName() const
    {
      return _configName;
    }

    /**
     * Returns the configured type of this device.
     */
    const QString& configType() const
    {
      return _configType;
    }

    /**
     * Returns the name of this device, as used in the song file.
     */
    const QString& name() const
    {
      return _name;
    }

    /**
     * Returns the type of this device, as used in the song file.
     */
    const QString& type() const
    {
      return _type;
    }

    /**
     * Returns the first unused pattern.
     */
    uint newPattern() const;

    /**
     * Returns the pattern with index @p index.  The pattern is created if it
     * does not already exist.
     */
    Pattern& operator[](uint index);

    /**
     * Returns the pattern with index @p index.  Unlike the non-const
     * []-operator, the pattern will not be created if it doesn't exist.
     */
    const Pattern& operator[](uint index) const;

    /**
     * Returns the amount of patterns in this device.
     */
    uint size() const;

    /**
     * Removes the pattern with index @p index.  If the pattern does not exist,
     * nothing happens.
     */
    void remove(uint index);

    /**
     * Returns the pattern index at the measure @p measure.
     */
    uint sequence(uint measure) const;

    /**
     * Sets the pattern at measure @p measure to @p pattern.
     * If @p pattern is 0, the measure will be cleared.
     */
    void setSequence(uint measure, uint pattern);

    /**
     * Clears the sequence of this device.
     */
    void clearSequence();

    /**
     * Returns this device's event list.
     */
    const EventList& events() const
    {
      return _events;
    }

    /**
     * Returns this device's event list.
     */
    EventList& events()
    {
      return _events;
    }

    /**
     * Returns the general volume for this device (0-127).
     */
    const uint volume() const
    {
      return _volume;
    }

    /**
     * Sets the general volume for this device (0-127).
     */
    void setVolume(uint volume);

  protected:

    friend class SanetrackerDoc;

    typedef QMap<uint, uint> Sequence;

    static const Pattern _emptyPattern;

    const Configuration& _config;

    uint                 _portIndex;
    uint                 _deviceID;
    QString              _configName;
    QString              _configType;

    QString              _name;
    QString              _type;

    QMap<uint, Pattern>  _patterns;
    Sequence             _sequence;
    EventList            _events;
    uint                 _volume;
  };

  /**
   * Deletes the contents of this document.  Returns TRUE on success, and
   * FALSE on failure.  If the call fails, use @see error to get an error
   * message.
   */
  bool newDocument();

  /**
   * Opens the document specified by @p url.  Returns TRUE if the opening
   * was successful, and FALSE otherwise.  If opening failed, call
   * @ref error to get an error message.
   */
  bool openDocument(const KURL& url);

  /**
   * Saves the document to the url specified by @p url.  Returns TRUE if the
   * saving was successful, and FALSE otherwise.  If saving failed, call
   * @ref error to get an error message.
   */
  bool saveDocument(const KURL& url);

  /**
   * Returns TRUE if the document has been modified since last save, and
   * FALSE otherwise.
   *
   * @see setModified
   */
  bool isModified() const
  {
    return _modified;
  };

  /**
   * Returns the error message of the last failed call.
   */
  QString error() const;

  /**
   * Returns the URL of this document, i.e. the URL last opened from, saved
   * to or set by @ref setURL.
   *
   * @see setURL
   */
  const KURL& URL() const
  {
    return _url;
  }

  /**
   * Sets the URL of this document to @p url.
   *
   * @see URL
   */
  void setURL(const KURL& url)
  {
    _url = url;
  }

  /**
   * Returns the song length in patterns (measures).
   */
  uint songLength() const;

  /**
   * Returns the device with the index @p index.
   */
  Device& operator[](uint index);

  /**
   * Returns the device with the index @p index.
   */
  const Device& operator[](uint index) const;

  /**
   * Returns the amount of devices.
   */
  uint size() const;

  /**
   * Inserts an empty measure in all devices before measure @p index.
   */
  void insertMeasure(uint index);

  /**
   * Removes the measure with index @p index from all devices.
   */
  void removeMeasure(uint column);

  /**
   * Returns the global event list for this song.  This structure should
   * not containt note events, controller events, or other events sent to
   * just one device at a time.
   */
  EventList& events() { return _events; }

  /**
   * Returns the global event list for this song.  This structure should
   * not containt note events, controller events, or other events sent to
   * just one device at a time.
   */
  const EventList& events() const { return _events; }

  /**
   * Returns the length in ticks of the measure with index @p index.
   */
  uint measureLength(uint index) const;


public slots:

  /**
   * Sets the modified-status of the document to @p modified.
   */
  void setModified(bool modified = true);

  /**
   * Reconfigures the document because of an altered configuration.
   */
  void reconfigure();

signals:

  void modified(bool modified);
  void volumeChanged();

protected:

  struct TimeBase
  {
    uint position;
    uint bpm;
    uint numerator;
    uint denominator;
    uint clocks;
    uint freq;

    TimeBase* next;
  };

  KURL                     _url;
  TimeBase                 _timeBase;

  const Configuration&     _config;

  bool _modified;

  EventList                _events;
  mutable QPtrList<Device> _devices;
  uint                     _ppqn; // pulses per quarter note

  QString                  _error;
};

#endif // SANETRACKERDOC_H

// vim:ts=2 sw=2 et
